import dataclasses
import enum
import typing

type Value = typing.Any

type ValueTree = dict[str, ValueTree] | Value
type Path = typing.Collection[str]
type Items = typing.Iterable[Item]


@dataclasses.dataclass
class Item:
    path: Path
    value: Value


class ConflictPolicy(enum.Enum):
    KEEP = enum.auto()
    OVERWRITE = enum.auto()
    NEST = enum.auto()
    TRIM = enum.auto()
    RAISE = enum.auto()


@dataclasses.dataclass
class Conflict:
    existing: ValueTree
    new: Item


@dataclasses.dataclass
class ConflictError(Exception):
    conflict: Conflict


def list_items(tree: ValueTree) -> Items:
    if not isinstance(tree, dict):
        return [Item([], tree)]

    items: list[Item] = []
    for k, v in tree.items():
        items.extend(Item([k, *i.path], i.value) for i in list_items(v))
    return items


def should_overwrite(conflict: Conflict, on_conflict: ConflictPolicy | None):
    match on_conflict:
        case ConflictPolicy.KEEP:
            return False
        case ConflictPolicy.OVERWRITE:
            return True
        case ConflictPolicy.NEST:
            return len(conflict.new.path) > 0
        case ConflictPolicy.TRIM:
            return len(conflict.new.path) == 0
        case _:
            raise ConflictError(conflict)


def set_item(tree: ValueTree, item: Item, on_conflict: ConflictPolicy | None) -> ValueTree:
    if len(item.path) == 0:
        if (
            not isinstance(tree, dict)
            or len(tree) == 0
            or should_overwrite(Conflict(tree, item), on_conflict)
        ):
            return item.value
        return tree

    if not isinstance(tree, dict):
        if not should_overwrite(Conflict(tree, item), on_conflict):
            return tree
        tree = {}

    key, *path = item.path

    try:
        tree[key] = set_item(tree.get(key, {}), Item(path, value=item.value), on_conflict)
    except ConflictError as e:
        e.add_note(f'at {key}')
        raise

    return tree



from .config import read_config
from .environment import Environment, InvalidValueError, NoEnvironmentError
from .item import Conflict, ConflictError, ConflictPolicy, Item, Items
from .layer import (
    Layer,
    Source,
    Transformer,
    case_transform,
    delimiter_transformer,
    env_layer,
    env_source,
    file_layer,
    file_source,
    item_transform,
    json_file_layer,
    prefix_transformer,
    toml_file_layer,
)

__all__ = [
    'read_config',
    'Environment',
    'InvalidValueError',
    'NoEnvironmentError',
    'Conflict',
    'ConflictError',
    'ConflictPolicy',
    'Item',
    'Items',
    'Layer',
    'Source',
    'Transformer',
    'case_transform',
    'delimiter_transformer',
    'env_layer',
    'env_source',
    'file_layer',
    'file_source',
    'item_transform',
    'json_file_layer',
    'prefix_transformer',
    'toml_file_layer',
]



import abc
import dataclasses
import io
import json
import os
import pathlib
import tomllib
import typing

from slcfg import item

type Transformer[T, S] = typing.Callable[[T], S]
type Layer = Source[item.Items]


@dataclasses.dataclass
class Source[T](abc.ABC):
    getter: typing.Callable[[], T]

    def __or__[S](self, t: Transformer[T, S]):
        return Source(getter=lambda: t(self.getter()))


### File Sources


def file_source(path: pathlib.Path, *, optional: bool):
    def getter():
        try:
            return io.BytesIO(path.read_bytes())
        except FileNotFoundError:
            if optional:
                return io.BytesIO()
            raise

    return Source(getter=getter)


def file_layer(
    path: pathlib.Path, parser: Transformer[io.BytesIO, typing.Any], *, optional: bool = False
):
    return file_source(path, optional=optional) | parser | item.list_items


def json_file_layer(path: pathlib.Path, *, optional: bool = False):
    return file_layer(path, json.load, optional=optional)


def toml_file_layer(path: pathlib.Path, *, optional: bool = False):
    return file_layer(path, tomllib.load, optional=optional)


### Env Sources


def env_source():
    return Source(getter=lambda: list(os.environ.items()))


def case_transform(vars: list[tuple[str, str]]):
    return [(k.lower(), v) for k, v in vars]


def prefix_transformer(prefix: str):
    def transormer(vars: list[tuple[str, str]]):
        return [(k.removeprefix(prefix), v) for k, v in vars if k.startswith(prefix)]

    return transormer


def delimiter_transformer(delimiter: str):
    def transormer(vars: list[tuple[str, str]]):
        return [(k.split(delimiter), v) for k, v in vars]

    return transormer


def item_transform(vars: list[tuple[list[str], str]]) -> item.Items:
    return [item.Item(k, v) for k, v in vars]


def env_layer(prefix: str, nested_delimiter: str, *, case_sensitive: bool = False):
    source = env_source()

    if not case_sensitive:
        prefix = prefix.lower()
        nested_delimiter = nested_delimiter.lower()
        source = source | case_transform

    return (
        source
        | prefix_transformer(prefix)
        | delimiter_transformer(nested_delimiter)
        | item_transform
    )



import typing

from slcfg import item as itemlib
from slcfg import layer as layerlib


def read_config[T](
    model: typing.Callable[..., T],
    layers: list[layerlib.Layer],
    on_conflict: itemlib.ConflictPolicy | None = None,
):
    value_tree: itemlib.ValueTree = {}
    for layer in layers:
        for item in layer.getter():
            value_tree = itemlib.set_item(value_tree, item, on_conflict)

    return model(**value_tree)



import enum
import os
import typing


class InvalidValueError(Exception):
    pass


class NoEnvironmentError(Exception):
    pass


class Environment(enum.Enum):
    @classmethod
    def get_from_env(
        cls, var_name: str, *, default: typing.Self | None = None, ignore_invalid: bool = False
    ) -> typing.Self:
        value = os.environ.get(var_name)
        if value:
            for self in cls:
                if self.name.lower() == value.lower():
                    return self

            missing_reason = (
                f"The value '{value}' for environment variable '{var_name}' is invalid, "
                f"possible choices are '{', '.join(s.name for s in cls)}"
            )
            if not ignore_invalid:
                raise InvalidValueError(missing_reason)
        else:
            missing_reason = f"Environment variable '{var_name}' is not defined"

        if default is not None:
            return default

        message = f'{missing_reason} and not default value was provided'
        raise NoEnvironmentError(message)
